;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; 02_stacks_4.lisp - Zásobníkové programování
;;;
;;; Jazyk pro myšlený zásobníkový stroj.
;;;


#|

Pomocné nástroje, hlavně práce se seznamy.
Definujeme jen ty, které budeme dále potřebovat.

Se znalostmi z minulých přednášek nepotřebují komentář.

|#

;; Slovo :cons: (car cdr -- (car . cdr))
(defprim :cons
  (word-bin-op #'cons))

;; Slovo :split: ((car . cdr) -- car cdr)
(defprim :split
  (let ((pair (pop *rslt*)))
    (push (car pair) *rslt*)
    (push (cdr pair) *rslt*)))

;; Slovo :pair?: (elem -- ano/ne) 
;; Testuje, zda je vrchol datového zásobníku pár.
(defprim :pair?
  (push (consp (pop *rslt*))
        *rslt*))

;; Procedura find na hledání prvku v seznamu --- (el list -- ano/ne)
(execute '(:dup :if                                ;el list
                   :drop :drop nil :else           ;list je prázdný
                                                   ;list je neprázdný
                   :split :rot- :over :swap        ;cdr el el car
                   :eql :if :swap                  ;el cdr
                            find :val :clsub :else ;když el se nerovná car
                            :drop :drop t :then    ;rovná se
                                   :then
           :ret)
         'find :bind)

#|
(execute 3 '(1 2 3) 'find :val :clsub)
(execute 4 '(1 2 3) 'find :val :clsub)
|#

;; Procedura append na spojení dvou seznamů: 
;; ((a1 ... am) (b1 ... bn) -- (a1 ... am b1 ... bn))
(execute '(:swap :dup :if                     ;(b1 ... bn) (a1 ... am)
                         :drop :else          ;m = 0
                         :split :rot          ;a1 (a2 ... am)  (b1 ... bn)
                         append :val :clsub   ;a1 (a2 ... am b1 ... bn)
                         :cons :then          ;(a1 ... am b1 ... bn)
           :ret)
         'append :bind)

#|
(execute '(1 2 3) '(4 5 6) 'append :val :clsub)
|#

;; Slovo :symbol?: (elem -- ano/ne) 
;; Testuje, zda je vrchol datového zásobníku symbol.
(defprim :symbol?
  (push (symbolp (pop *rslt*))
        *rslt*))

;; Slovo a procedura na tisk
;; Vytiskne hodnotu na datovém zásobníku a nechá ji tam

(execute '(:print :ret)
         'print :bind)

#|
(evaluate '(print (+ 1 1)))
|#



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; fáze 4 - interpretace vloženého Scheme
;;;

#|
Interpret definujeme čistě v našem jazyce, postupným vykonáním několika
programů. Většinou půjde o nastavení hodnoty proměnné na nějakou
novou proceduru.

Díky tomu, že máme naprogramovanou práci se seznamy, nebudeme už potřebovat 
žádná nová vestavěná slova.

Každou novou definici se okamžitě snažíme testovat.
|#

#|
Interpret budeme spouštět pomocnou funkcí Lispu evaluate, která vykoná
(funkcí execute) jednoduchý program, který uloží na datový zásobník
vyhodnocovaný výraz Scheme a pak spustí proceduru uloženou v proměnné
eval.

Rozlišujeme dva termíny:
- vykonání - jedná se o spuštění podprogramu zásobníkového jazyka
    slovem :clsub nebo celého programu funkcí execute
- vyhodnocení - interpretace výrazu Scheme procedurou uloženou 
    v proměnné eval nebo z Lispu funkcí evaluate, která to udělá přes
    funkci execute
|#

;; Pomocná funkce, spouští vyhodnocení výrazu Scheme tím,
;; že spustí (funkcí execute) vykonání jednoduchého programu.
(defun evaluate (expr)
  (execute expr 'eval :val :clsub))

;; Vyhodnocení obecného výrazu --- procedura eval
;; Definice nové procedury a její uložení do proměnné eval
;; Procedura vyhodnotí vrchol zásobníku rslt a uloží na něj výsledek
;; Konstanty a proměnné vyhodnocuje sama, na vyhodnocení složeného
;; výrazu použije proceduru v proměnné eval-comp.
(execute '(:dup :pair? :if :dup :symbol? 
                           :if :else             ;výraz je konstanta, neděláme nic
                             :val :then          ;symbol
                         :else
                           eval-comp :val :clsub ;složený výraz
                         :then
           :ret)
         'eval :bind)

#|
(evaluate 1)
(evaluate 'pi)
|#

#|
Složené výrazy vyhodnocujeme pomocí globální proměnné specials,
ve které je uložen seznam speciálních operátorů (symbolů).
Aby se výrazy se speciálním operátorem snadno vyhodnocovaly,
budeme ukládat procedurum která to dělá, přímo do operátoru.
Postupně tak vzniknou proměnné quote, if, lambda atd.
|#

;; Uložení seznamu speciálních operátorů do proměnné specials
(execute '(quote if lambda) 'specials :bind)

;; Vyhodnocení slož. výrazu --- eval-comp. 
;; ((op a1 ... an) -- hodnota)
(execute '(:split :swap              ;(a1 ... an) op
           :dup specials :val        ;(a1 ... an) op op (quote ...)
                find :val :clsub     ;(a1 ... an) op nil/t
                :if                  ;op není speciální
                                     ;spustíme pomocnou proceduru eval-app
                                     ;na vyhodnocení obecné aplikace
                  eval-app :val :clsub :else
                  :val :clsub :then  ;spustíme proceduru uloženou v op
           :ret)
         'eval-comp :bind)

;; Pomocná procedura leval
;; Vyhodnocení výrazů v seznamu, výsledky uloženy na rslt
;; ((e1 ... en) -- val1 ... valn)
(execute '(:dup :if :drop :else             ;seznam je prázdný, nevracíme nic
                    :split :swap            ;(e2 ... en) e1
                    eval :val :clsub        ;(e2 ... en) val1
                    :swap                   ;val1 (e2 ... en)
                    leval :val :clsub ;rekurze
                          :then
           :ret)
         'leval :bind)

#|
(execute '(5 pi 6 pi) 'leval :val :clsub)
(execute '(5 pi 6 pi) 'leval :val :clsub nil :cons :cons :cons :cons)

|#

#|
Aplikaci vyhodnotíme tak, že nejprve operátor připojíme k ostatním výrazům 
zezadu. Výraz (op a1 ... an) tedy převedeme na (a1 ... an op).
Pak použijeme pomocnou proceduru, která vyhodnotí všechny prvky seznamu
a hodnoty uloží za sebe na zásobník rslt. Hodnota operátoru tedy bude nahoře:
a1-val ... an-val op-val. Hodnoty jsou zorganizovány tak, že proceduru op-val
můžeme prostě spustit (argumenty jsou ve správném pořadí).

Pozor, důsledkem tohoto řešení je, že se v původním výrazu operátor vyhodnocuje
až jako poslední! (Ve Scheme to nevadí, ale může nás to plést v případě 
použití vedlejšího efektu). 
|#

;; Vyhodnocení aplikace --- eval-app.
;;((a1 ... an) op -- hodnota)
(execute '(nil :cons append :val :clsub  ;(a1 ... an op)
           leval :val :clsub             ;a1-val ... an-val op-val
           :clsub                        ;zavolání procedury op-val
           :ret)
         'eval-app :bind)


#|
(evaluate '(- 5 2))
(evaluate '(* (- 5 2) (+ 2 3)))

;; test pořadí vyhodnocování:
(evaluate '(print ((print /) (print 1) (print 2))))
|#

;; Teď postupně definujeme speciální operátory
;; Do názvů spec. operátorů ukládáme podprogramy, které
;; příslušný výraz vyhodnocují

;; Operátor quote. Vstup: (expr)
(execute '(:split :drop ;expr
           :ret)
         'quote :bind)

#|
(evaluate '(quote (xx yy zz)))
|#

;; Operátor if. Vstup: (test yes no)
(execute '(:split :split :split :drop  ;test yes no
           :rot                        ;yes no test
           eval :val :clsub            ;yes no test-val
           :if eval :val :clsub        ;yes no-val
               :swap :drop :else       ;no-val
               :drop                   ;yes
               eval :val :clsub :then  ;yes-val
           :ret)
         'if :bind)

#|
;; Pokud nejsou vyhodnocené definice dále, tak tady dojde
;; k chybě, protože není definována proměnná t:
(evaluate '(if t 2 3))

;; Takto to ale jde, protože máme spec. operátor quote:
(evaluate '(if (quote nil) 2 3))
;; Tohle je totéž (na (quote nil) to převede ještě Lisp):
(evaluate '(if 'nil 2 3))

(evaluate '(if (= 1 2) 2 3))
|#

#|
Hodnotou výrazu s operátorem lambda bude procedura zásobníkového
jazyka. Rekurzivně:

(lambda ()
  body)

se vyhodnotí na

(body eval :val :clsub :ret)

Dále:

(lambda (x)
  body)          -->     (x :bind (body eval :val :clsub :ret)
                                  :clsub
                            :unbind
                          :ret)

(lambda (x y)
  body)                  -->   (y :bind (x :bind (body eval :val :clsub :ret)
                                                 :clsub
                                           :unbind
                                       :ret)
                                       :clsub
                                  :unbind
                               :ret)

atd.

V těle operátoru lambda smí být jen jeden výraz!
|#

;; Operátor lambda. Vstup: ((x1 ... xn) body)
(execute '(:split :split :drop             ;(x1 ... xn) body
           (eval :val :clsub :ret) :cons   ;(x1 ... xn) (body eval :val :clsub :ret)
           xlambda :val :clsub             ;použití pomocné procedury
           :ret)
         'lambda :bind)

;; Pomocný podprogram xlambda. Vstup: (x1 ... xn) sub
(execute '(:swap                           ;sub (x1 ... xn)
           :dup :if                        ;n = 0
                  :drop :else              ;sub
                  :split :rot- :swap       ;(x2 ... xn) x1 sub
                  (:clsub :unbind :ret)    ;(x2 ... xn) x1 sub (:clsub :unbind :ret)
                    :cons                  ;(x2 ... xn) x1 (sub :clsub :unbind :ret)
                  (:bind) :swap            ;(x2 ... xn) x1 (:bind) (sub :clsub :unbind :ret)
                  append :val :clsub       ;(x2 ... xn) x1 (:bind sub :clsub :unbind :ret)
                    :cons                  ;(x2 ... xn) (x1 :bind sub :clsub :unbind :ret)
                  xlambda :val :clsub :then
            :ret)
         'xlambda :bind)

#|
(evaluate '(lambda () body))
(evaluate '(lambda (x) body))
(evaluate '(lambda (x y) body))

(evaluate '(lambda (x y) (+ x y)))
(evaluate '(lambda () 1))
(evaluate '((lambda () 1)))
(evaluate '((lambda (x y) (+ x y)) 2 3))
|#
                  


#|

;; Větší test:
(execute '(lambda (n)
            (if (= n 0)
                1
              (* n (fact (- n 1)))))
         'eval :val :clsub
         'fact :bind)

(evaluate '(fact 5))

|#

;; Procedury na práci se seznamy:

(execute '(:pair? :ret)
         'pair? :bind)

(execute '(:cons :ret)
         'cons :bind)

(execute '(:split :drop :ret)
         'car :bind)

(execute '(:split :swap :drop :ret)
         'cdr :bind)

#|
(evaluate '(cons (cons 1 (cons 2 3)) 4))
(evaluate '(car (cons 1 2)))
(evaluate '(cdr (cons 1 2)))
|#


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; fáze 5 - nové operátory vloženého Scheme
;;;

;; Nejprve nové speciální operátory define a begin.
;; Původní vazbu proměnné specials překrýváme, protože
;; zatím neumíme nastavovat hodnoty proměnných.
(execute 'define 'begin 'specials :val :cons :cons 'specials :bind)

#|
(execute 'specials :val)
|#

;; Operátor define slouží k definici (vytvoření vazby) nových proměnných.
;; Používá se samostatně (ne uvnitř jiného kódu) pro globální proměnné.
;; Syntax: (define var expr) => var
;; Vstup je tedy: (var expr)
(execute '(:split :split :drop   ;var expr
           eval :val :clsub      ;var expr-val
           :over                 ;var expr-val var
           :bind :ret)           ;var (a vytvoří se nová vazba)
         'define :bind)

(evaluate '(define nil 'nil))
(evaluate '(define t 't))

#|
(evaluate 'nil)
(evaluate 't)
(evaluate '(define sqr (lambda (n) (* n n))))
(evaluate '(sqr 5))
|#

#|
;; Test příkladu z 10. přednášky:

(evaluate '(define f (lambda (x)
                       ((g 'b)))))

(evaluate '(define g (lambda (x)
                       (lambda () x))))

;; a test:
(evaluate '(f 'a))

|#

#|
Teď už je náš Scheme zcela samostatný. Pokud nebudeme potřebovat
definovat nové primitivum nebo udělat něco velmi speciálního, 
můžeme vše dělat ve Scheme.
|#

#|
Ukážeme si to na novém speciálním operátoru begin, který definujeme
kompletně ve Scheme (níže je zakomentovaná definice v zásobníkovém 
jazyce)
|#

;; Operátor begin je obdobou operátoru progn z Lispu.
;; Je třeba zařídit, aby se ignorovaly hodnoty všech výrazů
;; kromě posledního.
;; Předpokládáme, že má vždy alespoň jeden argument.
(execute '(:split
           :dup :if
                   :drop eval :val :clsub :else
                   :swap 
                   eval :val :clsub
                   :drop
                   begin :val :clsub :then
           :ret)
         'begin :bind)


#|
(evaluate '(begin (print 1)
                  (print 2)
                  (print 3)))
|#


#|
;; Když máme operátor begin, můžeme vylepšit operátor lambda, aby 
;; v jeho těle šlo uvést více než jeden výraz. Třeba takto:

;; Operátor lambda. Vstup: ((x1 ... xn) body1 ... bodyn)
|#

(execute '(:split                          ;(x1 ... xn) (body1 ... bodyn)
           begin :swap :cons               ;(x1 ... xn) (begin body1 ... bodyn)
           (eval :val :clsub :ret) :cons   ;(x1 ... xn) ((begin body1 ... bodyn) 
                                           ;             eval :val :clsub :ret)
           xlambda :val :clsub             ;použití pomocné procedury
           :ret)
         'lambda :set!)



